"""
AA-Forum path config
"""

# Django
from django.urls import path

# AA Forum
from aa_forum.constants import INTERNAL_URL_PREFIX
from aa_forum.views import admin, forum, search

app_name: str = "aa_forum"

# IMPORTANT
# All internal URLs need to start with the designated prefix
# to prevent conflicts with user generated forum URLs

urlpatterns = [
    path("", forum.index, name="forum_index"),
    # Admin URLs
    path(f"{INTERNAL_URL_PREFIX}/admin/", admin.index, name="admin_index"),
    path(
        f"{INTERNAL_URL_PREFIX}/admin/category/create/",
        admin.category_create,
        name="admin_category_create",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/edit/",
        admin.category_edit,
        name="admin_category_edit",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/delete/",
        admin.category_delete,
        name="admin_category_delete",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/board/create/",
        admin.board_create,
        name="admin_board_create",
    ),
    path(
        (
            f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/board/"
            "<int:board_id>/create-child-board/"
        ),
        admin.board_create_child,
        name="admin_board_create_child",
    ),
    path(
        (
            f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/board/"
            "<int:board_id>/edit/"
        ),
        admin.board_edit,
        name="admin_board_edit",
    ),
    path(
        (
            f"{INTERNAL_URL_PREFIX}/admin/category/<int:category_id>/board/"
            "<int:board_id>/delete/"
        ),
        admin.board_delete,
        name="admin_board_delete",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/ajax/admin/category-order/",
        admin.ajax_category_order,
        name="admin_ajax_category_order",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/ajax/admin/board-order/",
        admin.ajax_board_order,
        name="admin_ajax_board_order",
    ),
    # Service URLs
    path(
        f"{INTERNAL_URL_PREFIX}/message/<int:message_id>/delete/",
        forum.message_delete,
        name="forum_message_delete",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/topic/<int:topic_id>/change-lock-state/",
        forum.topic_change_lock_state,
        name="forum_topic_change_lock_state",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/topic/<int:topic_id>/change-sticky-state/",
        forum.topic_change_sticky_state,
        name="forum_topic_change_sticky_state",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/topic/<int:topic_id>/delete/",
        forum.topic_delete,
        name="forum_topic_delete",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/search/",
        search.results,
        name="search_results",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/search/page/<int:page_number>/",
        search.results,
        name="search_results",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/unread/",
        forum.topic_show_all_unread,
        name="forum_topic_show_all_unread",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/mark-all-as-read/",
        forum.mark_all_as_read,
        name="forum_mark_all_as_read",
    ),
    # Forum URLs :: Board
    path(
        "<slug:category_slug>/<slug:board_slug>/",
        forum.board,
        name="forum_board",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/page/<int:page_number>/",
        forum.board,
        name="forum_board",
    ),
    path(
        f"{INTERNAL_URL_PREFIX}/<slug:category_slug>/<slug:board_slug>/new-topic/",
        forum.board_new_topic,
        name="forum_board_new_topic",
    ),
    # Forum URLs :: Topic
    path(
        "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/",
        forum.topic,
        name="forum_topic",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/"
        "<slug:topic_slug>/page/<int:page_number>/",
        forum.topic,
        name="forum_topic",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/reply/",
        forum.topic_reply,
        name="forum_topic_reply",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/unread/",
        forum.topic_first_unread_message,
        name="forum_topic_first_unread_message",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/modify/",
        forum.topic_modify,
        name="forum_topic_modify",
    ),
    # Forum URLs :: Message
    path(
        "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/<int:message_id>/",
        forum.message,
        name="forum_message",
    ),
    path(
        "<slug:category_slug>/<slug:board_slug>/"
        "<slug:topic_slug>/<int:message_id>/modify/",
        forum.message_modify,
        name="forum_message_modify",
    ),
]



"""
Test for our helper functions
"""

# Standard Library
from unittest.mock import patch

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_fake_user

# AA Forum
from aa_forum.forms import NewCategoryForm
from aa_forum.helper.eve_images import get_character_portrait_from_evecharacter
from aa_forum.helper.forms import message_form_errors
from aa_forum.helper.text import string_cleanup


@patch("aa_forum.helper.forms.messages")
class TestHelperForms(TestCase):
    """
    Testing the forms helper
    """

    def setUp(self) -> None:
        self.factory = RequestFactory()

    def test_should_send_form_errors_as_messages(self, messages):
        """
        Test should send form errors as message
        :param messages:
        :return:
        """

        # given
        form = NewCategoryForm({"name": "Dummy"})
        form.add_error("name", "error message 1")
        form.add_error(None, "error message 2")
        request = self.factory.get(reverse("aa_forum:forum_index"))

        # when
        message_form_errors(request, form)

        # then
        args, _ = messages.error.call_args_list[0]
        self.assertIn("error message 1", args[1])
        args, _ = messages.error.call_args_list[1]
        self.assertIn("error message 2", args[1])

    def test_should_do_nothing_when_form_has_no_errors(self, messages):
        """
        Test should do nothing when form has no errors
        :param messages:
        :return:
        """

        # given
        form = NewCategoryForm({"name": "Dummy"})
        request = self.factory.get(reverse("aa_forum:forum_index"))

        # when
        message_form_errors(request, form)

        # then
        self.assertFalse(messages.error.called)


class TestHelperText(TestCase):
    """
    Testing the text helper
    """

    def test_should_return_cleaned_string(self):
        """
        Test should return a clean/sanitized string
        :return:
        """

        # given
        string = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )

        # when
        cleaned_string = string_cleanup(string)

        # then
        self.assertIn(
            "this is a script test. and this is style test. end tests.", cleaned_string
        )


class TestHelperEveImages(TestCase):
    """
    Testing the eve_images helpers
    """

    def setUp(self) -> None:
        self.user_1001 = create_fake_user(
            1001, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )

    def test_should_return_character_portrait_url(self):
        """
        Test should return character portrait URL
        :return:
        """

        character = self.user_1001.profile.main_character

        portrait_url = get_character_portrait_from_evecharacter(character=character)
        expected_url = f"https://images.evetech.net/characters/{character.character_id}/portrait?size=32"  # , pylint: disable=line-too-long

        self.assertEqual(portrait_url, expected_url)

    def test_should_return_character_portrait_html(self):
        """
        Test should return character portrait HTML image tag
        :return:
        """

        character = self.user_1001.profile.main_character

        portrait_html = get_character_portrait_from_evecharacter(
            character=character, as_html=True
        )
        expected_url = f"https://images.evetech.net/characters/{character.character_id}/portrait?size=32"  # , pylint: disable=line-too-long
        expected_html = (
            '<img class="aa-forum-character-portrait img-rounded" '
            f'src="{expected_url}" alt="{character.character_name}" '
            'width="32" height="32">'
        )

        self.assertEqual(portrait_html, expected_html)



"""
Django admin declarations
"""

# Django
from django.contrib import admin
from django.utils.safestring import mark_safe

# AA Forum
from aa_forum.models import Board, Category, Topic


class BaseReadOnlyAdminMixin:
    """
    Base "Read Only" mixin for admin models
    """

    def has_add_permission(self, request):
        """
        Has add permissions
        :param request:
        :return:
        """

        return False

    def has_change_permission(self, request, obj=None):
        """
        Has change permissions
        :param request:
        :param obj:
        :return:
        """

        return False

    def has_delete_permission(self, request, obj=None):
        """
        Has delete permissions
        :param request:
        :param obj:
        :return:
        """

        return False


@admin.register(Category)
class CategoryAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Category admin
    """

    list_display = ("name", "slug", "_board_count")
    exclude = ("is_collapsible",)

    def _board_count(self, obj):
        """
        Return the board count per category
        :param obj:
        :return:
        """

        return obj.boards.count()


@admin.register(Board)
class BoardAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Board admin
    """

    list_display = (
        "name",
        "slug",
        "parent_board",
        "_groups",
        "category",
        "_topics_count",
    )

    def _groups(self, obj):
        """
        Return the groups this board is restricted to as list
        :param obj:
        :return:
        """

        groups = obj.groups.all()

        if groups.count() > 0:
            return mark_safe("<br>".join([group.name for group in groups]))

        return ""

    def _topics_count(self, obj):
        """
        Return the topics count per board
        :param obj:
        :return:
        """

        return obj.topics.count()


@admin.register(Topic)
class TopicAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Topic admin
    """

    list_display = ("subject", "slug", "board", "_messages_count")

    def _messages_count(self, obj):
        """
        Return the message count per topic
        :param obj:
        :return:
        """

        return obj.messages.count()



"""
Test for the managers
"""

# Django
from django.contrib.auth.models import Group
from django.test import TestCase

# AA Forum
from aa_forum.models import Board, Category, Topic
from aa_forum.tests.utils import create_fake_message, create_fake_user


class TestBoard(TestCase):
    """
    Test board manager
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")
        cls.category = Category.objects.create(name="Science")

    def setUp(self) -> None:
        self.user = create_fake_user(1001, "Bruce Wayne")

    def test_should_return_board_with_no_groups(self):
        """
        Test should return biard with no groups
        :return:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)

        # when
        result = Board.objects.user_has_access(self.user)

        # then
        self.assertIn(board, result)

    def test_should_return_board_for_group_member(self):
        """
        Test should return board for group member
        :return:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)
        board.groups.add(self.group)
        self.user.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(self.user)

        # then
        self.assertIn(board, result)

    def test_should_not_return_board_for_non_group_member(self):
        """
        Test should not return board for non group member
        :return:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)
        board.groups.add(self.group)

        # when
        result = Board.objects.user_has_access(self.user)

        # then
        self.assertNotIn(board, result)


class TestTopic(TestCase):
    """
    Tests for topic manager
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.group = Group.objects.create(name="Superhero")
        cls.category = Category.objects.create(name="Science")

    def setUp(self) -> None:
        self.user = create_fake_user(1001, "Bruce Wayne")
        self.board = Board.objects.create(name="Physics", category=self.category)

    def test_should_return_topic_normally_1(self):
        """
        Test should return topic normally
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=str(self.category.slug),
            board_slug=str(self.board.slug),
            topic_slug=str(topic.slug),
            user=self.user,
        )

        # then
        self.assertEqual(result, topic)

    def test_should_return_topic_normally_2(self):
        """
        Test should return topic normally (test without positional arguments)
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            self.category.slug,
            self.board.slug,
            topic.slug,
            self.user,
        )

        # then
        self.assertEqual(result, topic)

    def test_should_return_none_if_not_found_1(self):
        """
        Test should return None if not found (invalid topic slug)
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug=self.board.slug,
            topic_slug="invalid",
            user=self.user,
        )

        # then
        self.assertIsNone(result)

    def test_should_return_none_if_not_found_2(self):
        """
        Test should return None if not found (invalid board slug)
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug=self.category.slug,
            board_slug="invalid",
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertIsNone(result)

    def test_should_return_none_if_not_found_3(self):
        """
        Test should return None if not found (invalid category slug)
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            category_slug="invalid",
            board_slug=self.board.slug,
            topic_slug=topic.slug,
            user=self.user,
        )

        # then
        self.assertIsNone(result)

    def test_should_return_restricted_topic_for_group_members(self):
        """
        Test should return restricted topic for group member
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        self.user.groups.add(self.group)
        self.board.groups.add(self.group)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            self.category.slug,
            self.board.slug,
            topic.slug,
            self.user,
        )

        # then
        self.assertEqual(result, topic)

    def test_should_not_return_restricted_topic_for_non_group_members(self):
        """
        Test should not return restricted topic for non group member
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        self.board.groups.add(self.group)
        create_fake_message(topic, self.user)

        # when
        result = Topic.objects.get_from_slugs(
            self.category.slug,
            self.board.slug,
            topic.slug,
            self.user,
        )

        # then
        self.assertIsNone(result)



"""
Set custom variables in a Django template
"""

# Django
from django import template

register = template.Library()


class SetVarNode(template.Node):
    """
    SetVarNode
    """

    def __init__(self, var_name, var_value):
        self.var_name = var_name
        self.var_value = var_value

    def render(self, context):
        """
        Render context
        :param context:
        :return:
        """

        try:
            value = template.Variable(self.var_value).resolve(context)
        except template.VariableDoesNotExist:
            value = ""

        context[self.var_name] = value

        return ""


@register.tag(name="set_template_variable")
def set_template_variable(parser, token):
    """
    Set a template variable
    Usage: {% set_template_variable <var_name> = <var_value> %}
    :param parser:
    :param token:
    :return:
    """

    parts = token.split_contents()

    if len(parts) < 4:
        raise template.TemplateSyntaxError(
            "'set_template_variable' tag must be of the form: "
            "{% set_template_variable <var_name> = <var_value> %}"
        )

    return SetVarNode(parts[1], parts[3])
